# Copyright (c) 2021 Ole-Christoffer Granmo

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This code implements the Convolutional Tsetlin Machine from paper arXiv:1905.09688
# https://arxiv.org/abs/1905.09688

import numpy as np

import tsetlinmachinecuda.kernels as kernels

import pycuda.curandom as curandom
import pycuda.driver as cuda
from pycuda.compiler import SourceModule

from time import time

cuda.init()

class GPU():
	None

class CommonTsetlinMachine():
	def __init__(self, number_of_clauses, T, s, clause_drop_p=0.0, feature_drop_p=0.0, number_of_gpus=1, q=1.0, boost_true_positive_feedback=1, weighted_clauses=0, number_of_state_bits=8, append_negated=True, grid=(16*13,1,1), block=(128,1,1)):
		self.number_of_gpus = np.minimum(cuda.Device.count(), number_of_gpus)

		self.number_of_clauses = number_of_clauses
		self.number_of_clauses_multi = int(number_of_clauses // self.number_of_gpus)
		self.number_of_state_bits = number_of_state_bits
		self.T = int(T)
		self.s = s
		self.q = q
		self.boost_true_positive_feedback = boost_true_positive_feedback
		self.weighted_clauses = weighted_clauses
		self.append_negated = append_negated
		self.grid = grid
		self.block = block
		
		self.clause_drop_p = clause_drop_p
		self.feature_drop_p = feature_drop_p

		self.X_train = np.array([])
		self.Y_train = np.array([])
		self.X_test = np.array([])
		self.ta_state = np.array([])
		self.clause_weights = np.array([])

		self.initialized = False

		self.gpus = []
		for c in range(self.number_of_gpus):
			print("Preparing GPU #%d" % (c))
			gpu = GPU()
			gpu.device_id = c
			gpu.device = cuda.Device(c)
			gpu.context = gpu.device.make_context()
			gpu.g = curandom.XORWOWRandomNumberGenerator() 

			gpu.mod_encode = SourceModule(kernels.code_encode, no_extern_c=True)
			gpu.prepare_encode = gpu.mod_encode.get_function("prepare_encode")
			gpu.encode = gpu.mod_encode.get_function("encode")

			self.gpus.append(gpu)

			gpu.context.pop()
		print()

	def encode_X_multi(self, X, encoded_X_gpus):
		number_of_examples = X.shape[0]

		Xm = np.ascontiguousarray(X.flatten()).astype(np.uint32)

		X_gpus = []
		for gpu in self.gpus:
			gpu.context.push()
			encoded_X_gpu = encoded_X_gpus[gpu.device_id-1]
			X_gpus.append(cuda.mem_alloc(Xm.nbytes))
			cuda.memcpy_htod_async(X_gpus[-1], Xm)
			gpu.context.pop()

		for i in range(len(self.gpus)):
			gpu = self.gpus[i]

			gpu.context.push()
			gpu.context.synchronize()
			
			encoded_X_gpu = encoded_X_gpus[i]
			X_gpu = X_gpus[i]

			if self.append_negated:			
				gpu.prepare_encode(X_gpu, encoded_X_gpu, np.int32(number_of_examples), np.int32(self.dim[0]), np.int32(self.dim[1]), np.int32(self.dim[2]), np.int32(self.patch_dim[0]), np.int32(self.patch_dim[1]), np.int32(1), np.int32(0), grid=self.grid, block=self.block)
			else:
				gpu.prepare_encode(X_gpu, encoded_X_gpu, np.int32(number_of_examples), np.int32(self.dim[0]), np.int32(self.dim[1]), np.int32(self.dim[2]), np.int32(self.patch_dim[0]), np.int32(self.patch_dim[1]), np.int32(0), np.int32(0), grid=self.grid, block=self.block)
			gpu.context.pop()

		for i in range(len(self.gpus)):
			gpu = self.gpus[i]

			gpu.context.push()
			gpu.context.synchronize()
			
			encoded_X_gpu = encoded_X_gpus[i]
			X_gpu = X_gpus[i]

			if self.append_negated:			
				gpu.encode(X_gpu, encoded_X_gpu, np.int32(number_of_examples), np.int32(self.dim[0]), np.int32(self.dim[1]), np.int32(self.dim[2]), np.int32(self.patch_dim[0]), np.int32(self.patch_dim[1]), np.int32(1), np.int32(0), grid=self.grid, block=self.block)
			else:
				gpu.encode(X_gpu, encoded_X_gpu, np.int32(number_of_examples), np.int32(self.dim[0]), np.int32(self.dim[1]), np.int32(self.dim[2]), np.int32(self.patch_dim[0]), np.int32(self.patch_dim[1]), np.int32(0), np.int32(0), grid=self.grid, block=self.block)
			gpu.context.pop()

		self.sync_gpus() ## NOT NEEDED

	def allocate_gpu_memory_multi(self, number_of_examples):
		for gpu in self.gpus:
			gpu.context.push()
			gpu.ta_state_gpu = cuda.mem_alloc(self.number_of_clauses_multi*self.number_of_ta_chunks*self.number_of_state_bits*4)
			gpu.clause_weights_gpu = cuda.mem_alloc(self.number_of_clauses_multi*4)
			gpu.class_sum_gpu = cuda.mem_alloc(self.number_of_classes*4)
			gpu.drop_clause_gpu = cuda.mem_alloc(self.number_of_clauses_multi*4)
			gpu.drop_literal_gpu = cuda.mem_alloc(self.number_of_ta_chunks*4)
			gpu.context.pop()


	def get_weight(self, mc_tm_class, clause):
		global_clause = mc_tm_class*self.number_of_clauses//self.number_of_classes + clause
		gpu_clause = global_clause % self.number_of_clauses_multi
		gpu_id = global_clause//self.number_of_clauses_multi

		return(self.get_state()[1][gpu_id][gpu_clause])

	def ta_action(self, mc_tm_class, clause, ta):
		state = self.get_state()[0]
		global_clause = mc_tm_class*self.number_of_clauses//self.number_of_classes + clause
		gpu_clause = global_clause % self.number_of_clauses_multi
		gpu_id = global_clause//self.number_of_clauses_multi
		ta_state = state[gpu_id].reshape((self.number_of_clauses_multi, self.number_of_ta_chunks, self.number_of_state_bits))

		return (ta_state[gpu_clause, ta // 32, self.number_of_state_bits-1] & (1 << (ta % 32))) > 0

	def get_state(self):
		if np.array_equal(self.clause_weights, np.array([])):
			self.ta_state = []
			self.clause_weights = []
			for gpu in self.gpus:
				gpu.context.push()
				gpu.ta_state = np.empty(self.number_of_clauses_multi*self.number_of_ta_chunks*self.number_of_state_bits).astype(np.uint32)
				cuda.memcpy_dtoh(gpu.ta_state, gpu.ta_state_gpu)
				gpu.clause_weights = np.empty(self.number_of_clauses_multi).astype(np.int32)
				cuda.memcpy_dtoh(gpu.clause_weights, gpu.clause_weights_gpu)
				gpu.context.pop()

				self.ta_state.append(gpu.ta_state)
				self.clause_weights.append(gpu.clause_weights)
		return((self.ta_state, self.clause_weights, self.number_of_classes, self.number_of_clauses, self.number_of_features, self.dim, self.patch_dim, self.number_of_patches, self.number_of_state_bits, self.number_of_ta_chunks, self.append_negated, self.min_y, self.max_y))

	def set_state(self, state):
		self.number_of_classes = state[2]
		self.number_of_clauses = state[3]
		self.number_of_clauses_multi = int(self.number_of_clauses // self.number_of_gpus)
		self.number_of_features = state[4]
		self.dim = state[5]
		self.patch_dim = state[6]
		self.number_of_patches = state[7]
		self.number_of_state_bits = state[8]
		self.number_of_ta_chunks = state[9]
		self.append_negated = state[10]
		self.min_y = state[11]
		self.max_y = state[12]
		
		for i in range(len(self.gpus)):
			gpu = self.gpus[i]
			gpu.context.push()
			gpu.ta_state_gpu = cuda.mem_alloc(self.number_of_clauses_multi*self.number_of_ta_chunks*self.number_of_state_bits*4)
			gpu.clause_weights_gpu = cuda.mem_alloc(self.number_of_clauses_multi*4)
			cuda.memcpy_htod(gpu.ta_state_gpu, state[0][i])
			cuda.memcpy_htod(gpu.clause_weights_gpu, state[1][i])
			gpu.context.pop()

		self.X_train = np.array([])
		self.Y_train = np.array([])
		self.X_test = np.array([])
		self.ta_state = np.array([])
		self.clause_weights = np.array([])

	def sync_gpus(self):
		for gpu in self.gpus:
			gpu.context.push()
			gpu.context.synchronize()
			gpu.context.pop()

	# Transform input data for processing at next layer
	def transform(self, X):
		number_of_examples = X.shape[0]
		
		self.encoded_X_transform_gpus = []
		for gpu in self.gpus:
			gpu.context.push()
			self.encoded_X_transform_gpus.append(cuda.mem_alloc(int(number_of_examples * self.number_of_patches * self.number_of_ta_chunks*4)))
			gpu.context.pop()

		self.encode_X_multi(X, self.encoded_X_transform_gpus)

		parameters_multi = """
#define CLASSES %d
#define CLAUSES %d
#define FEATURES %d
#define STATE_BITS %d
#define BOOST_TRUE_POSITIVE_FEEDBACK %d
#define WEIGHTED_CLAUSES %d
#define S %f
#define THRESHOLD %d

#define NEGATIVE_CLAUSES %d

#define PATCHES %d

#define NUMBER_OF_EXAMPLES %d
		""" % (self.number_of_classes, self.number_of_clauses_multi, self.number_of_features, self.number_of_state_bits, self.boost_true_positive_feedback, self.weighted_clauses, self.s, self.T, self.negative_clauses, self.number_of_patches, number_of_examples)

		for gpu in self.gpus:
			gpu.context.push()
			mod = SourceModule(parameters_multi + kernels.code_header + kernels.code_transform, no_extern_c=True)
			gpu.transform = mod.get_function("transform")
			gpu.X_transformed_gpu = cuda.mem_alloc(number_of_examples*self.number_of_clauses_multi*4)
			gpu.context.pop()

		for i in range(len(self.gpus)):
			gpu = self.gpus[i]
			gpu.context.push()
			gpu.transform(gpu.ta_state_gpu, self.encoded_X_transform_gpus[i], gpu.X_transformed_gpu, grid=self.grid, block=self.block)
			gpu.context.pop()

		global_X_transformed = np.empty((number_of_examples, self.number_of_clauses), dtype=np.uint32)
		for i in range(len(self.gpus)):
			gpu = self.gpus[i]
			gpu.context.push()
			gpu.context.synchronize()
			local_X_transformed = np.ascontiguousarray(np.empty(number_of_examples*self.number_of_clauses_multi, dtype=np.uint32))
			cuda.memcpy_dtoh(local_X_transformed, gpu.X_transformed_gpu)
			global_X_transformed[:,i*self.number_of_clauses_multi:(i+1)*self.number_of_clauses_multi] = local_X_transformed.reshape((number_of_examples, self.number_of_clauses_multi))
			gpu.context.pop()

		return global_X_transformed

	def _fit(self, X, encoded_Y, epochs=100, incremental=False):
		number_of_examples = X.shape[0]

		if (not self.initialized):
			self.initialized = True
						
			if len(X.shape) == 3:
				self.dim = (X.shape[1], X.shape[2],  1)
			elif len(X.shape) == 4:
				self.dim = X.shape[1:]

			if self.append_negated:
				self.number_of_features = int(self.patch_dim[0]*self.patch_dim[1]*self.dim[2] + (self.dim[0] - self.patch_dim[0]) + (self.dim[1] - self.patch_dim[1]))*2
			else:
				self.number_of_features = int(self.patch_dim[0]*self.patch_dim[1]*self.dim[2] + (self.dim[0] - self.patch_dim[0]) + (self.dim[1] - self.patch_dim[1]))

			self.number_of_patches = int((self.dim[0] - self.patch_dim[0] + 1)*(self.dim[1] - self.patch_dim[1] + 1))
			self.number_of_ta_chunks = int((self.number_of_features-1)/32 + 1)
		
			parameters_multi = """
	#define CLASSES %d
	#define CLAUSES %d
	#define FEATURES %d
	#define STATE_BITS %d
	#define BOOST_TRUE_POSITIVE_FEEDBACK %d
	#define WEIGHTED_CLAUSES %d
	#define S %f
	#define THRESHOLD %d
	#define Q %f
	
	#define NEGATIVE_CLAUSES %d

	#define PATCHES %d

	#define NUMBER_OF_EXAMPLES %d
""" % (self.number_of_classes, self.number_of_clauses_multi, self.number_of_features, self.number_of_state_bits, self.boost_true_positive_feedback, self.weighted_clauses, self.s, self.T, self.q, self.negative_clauses, self.number_of_patches, number_of_examples)

			for gpu in self.gpus:
				gpu.context.push()
				mod_prepare = SourceModule(parameters_multi + kernels.code_header + kernels.code_prepare, no_extern_c=True)
				gpu.prepare = mod_prepare.get_function("prepare")
				gpu.context.pop()

			self.allocate_gpu_memory_multi(number_of_examples)

			for gpu in self.gpus:
				gpu.context.push()
				
				new_mod_update = SourceModule(parameters_multi + kernels.code_header + kernels.code_update, no_extern_c=True)
				gpu.update = new_mod_update.get_function("update")
				gpu.update.prepare("PPPPPPPPi")

				gpu.evaluate_update = new_mod_update.get_function("evaluate")
				gpu.evaluate_update.prepare("PPPPPPi")

				gpu.context.pop()

			self.encoded_X_training_gpus = []
			for gpu in self.gpus:
				gpu.context.push()
				self.encoded_X_training_gpus.append(cuda.mem_alloc(int(number_of_examples * self.number_of_patches * self.number_of_ta_chunks*4)))
				gpu.context.pop()

			self.Y_gpus = []
			for gpu in self.gpus:
				gpu.context.push()
				self.Y_gpus.append(cuda.mem_alloc(encoded_Y.nbytes))
				gpu.context.pop()

			for gpu in self.gpus:
				gpu.context.push()
				gpu.prepare(gpu.g.state, gpu.ta_state_gpu, gpu.clause_weights_gpu, gpu.class_sum_gpu, grid=self.grid, block=self.block)
				gpu.context.pop()

			self.sync_gpus()

		if (not np.array_equal(self.X_train, X)) or (not np.array_equal(self.encoded_Y_train, encoded_Y)):
			self.X_train = X
			self.encoded_Y_train = encoded_Y

			self.encode_X_multi(X, self.encoded_X_training_gpus)

			for i in range(len(self.gpus)):
				gpu = self.gpus[i]
				gpu.context.push()
				cuda.memcpy_htod_async(self.Y_gpus[i], encoded_Y)
				gpu.context.pop()

			self.sync_gpus()

		if incremental == False:
			for gpu in self.gpus:
				gpu.context.push()
				gpu.prepare(gpu.g.state, gpu.ta_state_gpu, gpu.clause_weights_gpu, gpu.class_sum_gpu, grid=self.grid, block=self.block)
				gpu.context.pop()
			self.sync_gpus()

		for epoch in range(epochs):
			drop_literal = (np.random.rand(self.number_of_features) <= self.feature_drop_p).astype(np.uint32)
			drop_literal_chunk = np.zeros(self.number_of_ta_chunks).astype(np.uint32)
			for k in range(self.number_of_features):
				if drop_literal[k] == 1:
					ta_chunk = k // 32
					ta_pos = k % 32
					drop_literal_chunk[ta_chunk] |= (1 << ta_pos)

			for i in range(len(self.gpus)):
				gpu = self.gpus[i]
				gpu.context.push()
				drop_clause = np.ascontiguousarray(np.random.rand(self.number_of_clauses_multi) <= self.clause_drop_p).astype(np.uint32)
				cuda.memcpy_htod(gpu.drop_clause_gpu, drop_clause)
				cuda.memcpy_htod(gpu.drop_literal_gpu, drop_literal_chunk)
				gpu.context.pop()

			for e in range(number_of_examples):
				for i in range(len(self.gpus)):
					gpu = self.gpus[i]
					gpu.context.push()
					class_sum = np.ascontiguousarray(np.zeros(self.number_of_classes)).astype(np.int32)
					cuda.memcpy_htod(gpu.class_sum_gpu, class_sum)

					gpu.evaluate_update.prepared_call(self.grid, self.block, gpu.ta_state_gpu, gpu.clause_weights_gpu, gpu.class_sum_gpu, gpu.drop_clause_gpu, gpu.drop_literal_gpu, self.encoded_X_training_gpus[i], np.int32(e))
					gpu.context.pop()

				global_class_sum = np.ascontiguousarray(np.zeros(self.number_of_classes)).astype(np.int32)
				for i in range(len(self.gpus)):
					gpu = self.gpus[i]
					gpu.context.push()
					gpu.context.synchronize()
					local_class_sum = np.ascontiguousarray(np.zeros(self.number_of_classes)).astype(np.int32)
					cuda.memcpy_dtoh(local_class_sum, gpu.class_sum_gpu)
					global_class_sum += local_class_sum
					gpu.context.pop()

				for i in range(len(self.gpus)):
					gpu = self.gpus[i]
					gpu.context.push()
					cuda.memcpy_htod(gpu.class_sum_gpu, global_class_sum)
					gpu.context.pop()

				for i in range(len(self.gpus)):
					gpu = self.gpus[i]
					gpu.context.push()
					gpu.context.synchronize()
					gpu.update.prepared_call(self.grid, self.block, gpu.g.state, gpu.ta_state_gpu, gpu.clause_weights_gpu, gpu.class_sum_gpu, gpu.drop_clause_gpu, gpu.drop_literal_gpu, self.encoded_X_training_gpus[i], self.Y_gpus[i], np.int32(e))
					gpu.context.pop()

				self.sync_gpus()

		self.ta_state = np.array([])
		self.clause_weights = np.array([])
		
		return

	def _score(self, X):
		number_of_examples = X.shape[0]
		
		if not np.array_equal(self.X_test, X):
			self.X_test = X

			self.encoded_X_test_gpus = []
			for gpu in self.gpus:
				gpu.context.push()
				self.encoded_X_test_gpus.append(cuda.mem_alloc(int(number_of_examples * self.number_of_patches * self.number_of_ta_chunks*4)))
				gpu.context.pop()

			self.encode_X_multi(X, self.encoded_X_test_gpus)

			parameters_multi = """
	#define CLASSES %d
	#define CLAUSES %d
	#define FEATURES %d
	#define STATE_BITS %d
	#define BOOST_TRUE_POSITIVE_FEEDBACK %d
	#define WEIGHTED_CLAUSES %d
	#define S %f
	#define THRESHOLD %d

	#define NEGATIVE_CLAUSES %d

	#define PATCHES %d

	#define NUMBER_OF_EXAMPLES %d
			""" % (self.number_of_classes, self.number_of_clauses_multi, self.number_of_features, self.number_of_state_bits, self.boost_true_positive_feedback, self.weighted_clauses, self.s, self.T, self.negative_clauses, self.number_of_patches, number_of_examples)

			self.class_sum_gpus = []
			for gpu in self.gpus:
				gpu.context.push()
				mod = SourceModule(parameters_multi + kernels.code_header + kernels.code_evaluate, no_extern_c=True)
				gpu.evaluate = mod.get_function("evaluate")
				local_class_sum = np.ascontiguousarray(np.empty(self.number_of_classes*number_of_examples)).astype(np.int32)
				self.class_sum_gpus.append(cuda.mem_alloc(local_class_sum.nbytes))
				gpu.context.pop()

		global_class_sum = np.ascontiguousarray(np.zeros(self.number_of_classes*number_of_examples)).astype(np.int32)
		for i in range(len(self.gpus)):
			gpu = self.gpus[i]
			gpu.context.push()
			local_class_sum = np.ascontiguousarray(np.zeros(self.number_of_classes*number_of_examples)).astype(np.int32)
			cuda.memcpy_htod(self.class_sum_gpus[i], local_class_sum)
			gpu.evaluate(gpu.ta_state_gpu, gpu.clause_weights_gpu, self.class_sum_gpus[i], self.encoded_X_test_gpus[i], grid=self.grid, block=self.block)
			gpu.context.pop()

		for i in range(len(self.gpus)):
			gpu = self.gpus[i]
			gpu.context.push()
			gpu.context.synchronize()
			cuda.memcpy_dtoh(local_class_sum, self.class_sum_gpus[i])
			global_class_sum += local_class_sum
			gpu.context.pop()
		
		class_sum_multi = np.clip(global_class_sum.reshape((self.number_of_classes, number_of_examples)), -self.T, self.T)

		return class_sum_multi
	
class MultiClassConvolutionalTsetlinMachine2D(CommonTsetlinMachine):
	"""
	This class ...
	"""
	
	def __init__(self, number_of_clauses, T, s, patch_dim, clause_drop_p=0.0, feature_drop_p=0.0, number_of_gpus=1, q=1.0, boost_true_positive_feedback=1, weighted_clauses=0, number_of_state_bits=8, append_negated=True, grid=(16*13,1,1), block=(128,1,1)):
		super().__init__(number_of_clauses, T, s, clause_drop_p=clause_drop_p, feature_drop_p=feature_drop_p, number_of_gpus=number_of_gpus, q=q, boost_true_positive_feedback=boost_true_positive_feedback, weighted_clauses=weighted_clauses, number_of_state_bits=number_of_state_bits, append_negated=append_negated, grid=grid, block=block)
		self.patch_dim = patch_dim
		self.negative_clauses = 1

	def fit(self, X, Y, epochs=100, incremental=False):
		self.number_of_classes = int(np.max(Y) + 1)
	
		self.max_y = None
		self.min_y = None
		
		encoded_Y = np.empty((Y.shape[0], self.number_of_classes), dtype = np.int32)
		for i in range(self.number_of_classes):
			encoded_Y[:,i] = np.where(Y == i, self.T, -self.T)

		self._fit(X, encoded_Y, epochs=epochs, incremental=incremental)

	def score(self, X):
		return self._score(X)

	def predict(self, X):
		return np.argmax(self.score(X), axis=0)

class MultiClassTsetlinMachine(CommonTsetlinMachine):
	def __init__(self, number_of_clauses, T, s, clause_drop_p=0.0, feature_drop_p=0.0, number_of_gpus=1, boost_true_positive_feedback=1, weighted_clauses=0, number_of_state_bits=8, append_negated=True, grid=(16*13,1,1), block=(128,1,1)):
		super().__init__(number_of_clauses, T, s, clause_drop_p=clause_drop_p, feature_drop_p=feature_drop_p, number_of_gpus=number_of_gpus, boost_true_positive_feedback=boost_true_positive_feedback, weighted_clauses=weighted_clauses, number_of_state_bits=number_of_state_bits, append_negated=append_negated, grid=grid, block=block)
		self.negative_clauses = 1

	def fit(self, X, Y, epochs=100, incremental=False):
		X = X.reshape(X.shape[0], X.shape[1], 1)

		self.number_of_classes = int(np.max(Y) + 1)
		self.patch_dim = (X.shape[1], 1, 1)

		self.max_y = None
		self.min_y = None
		
		encoded_Y = np.empty((Y.shape[0], self.number_of_classes), dtype = np.int32)
		for i in range(self.number_of_classes):
			encoded_Y[:,i] = np.where(Y == i, self.T, -self.T)

		self._fit(X, encoded_Y, epochs = epochs, incremental = incremental)

		return

	def score(self, X):
		X = X.reshape(X.shape[0], X.shape[1], 1)
		return self._score(X)

	def predict(self, X):
		return np.argmax(self.score(X), axis=0)

class TsetlinMachine(CommonTsetlinMachine):
	def __init__(self, number_of_clauses, T, s, clause_drop_p=0.0, feature_drop_p=0.0, number_of_gpus=1, boost_true_positive_feedback=1, weighted_clauses=0, number_of_state_bits=8, append_negated=True, grid=(16*13,1,1), block=(128,1,1)):
		super().__init__(number_of_clauses, T, s, clause_drop_p=clause_drop_p, feature_drop_p=feature_drop_p, number_of_gpus=number_of_gpus, boost_true_positive_feedback=boost_true_positive_feedback, weighted_clauses=weighted_clauses, number_of_state_bits=number_of_state_bits, append_negated=append_negated, grid=grid, block=block)
		self.negative_clauses = 1

	def fit(self, X, Y, epochs=100, incremental=False):
		X = X.reshape(X.shape[0], X.shape[1], 1)

		self.number_of_classes = 1
		self.patch_dim = (X.shape[1], 1, 1)
		
		self.max_y = None
		self.min_y = None
		
		encoded_Y = np.where(Y == 1, self.T, -self.T).astype(np.int32)

		self._fit(X, encoded_Y, epochs = epochs, incremental = incremental)

		return

	def score(self, X):
		X = X.reshape(X.shape[0], X.shape[1], 1)
		return self._score(X)[0,:]

	def predict(self, X):
		return self.score(X) >= 0

class RegressionTsetlinMachine(CommonTsetlinMachine):
	def __init__(self, number_of_clauses, T, s, number_of_gpus=1, boost_true_positive_feedback=1, weighted_clauses=0, number_of_state_bits=8, append_negated=True, grid=(16*13,1,1), block=(128,1,1)):
		super().__init__(number_of_clauses, T, s, clause_drop_p=clause_drop_p, feature_drop_p=feature_drop_p, number_of_gpus=number_of_gpus, boost_true_positive_feedback=boost_true_positive_feedback, weighted_clauses=weighted_clauses, number_of_state_bits=number_of_state_bits, append_negated=append_negated, grid=grid, block=block)
		self.negative_clauses = 0

	def fit(self, X, Y, epochs=100, incremental=False):
		X = X.reshape(X.shape[0], X.shape[1], 1)
		
		self.number_of_classes = 1
		self.patch_dim = (X.shape[1], 1, 1)

		self.max_y = np.max(Y)
		self.min_y = np.min(Y)
	
		encoded_Y = ((Y - self.min_y)/(self.max_y - self.min_y)*self.T).astype(np.int32)
			
		self._fit(X, encoded_Y, epochs = epochs, incremental = incremental)

		return

	def predict(self, X):
		X = X.reshape(X.shape[0], X.shape[1], 1)
		
		return 1.0*(self._score(X)[0,:])*(self.max_y - self.min_y)/(self.T) + self.min_y



from setuptools import *

setup(
   name='tsetlinmachinecuda',
   version='0.1.8',
   author='Ole-Christoffer Granmo',
   author_email='ole.granmo@uia.no',
   url='https://github.com/cair/tsetlinmachinecuda/',
   license='MIT',
   description='Massively Parallel and Asynchronous Architecture for Logic-based AI.',
   long_description='Massively Parallel and Asynchronous Architecture for Logic-based AI. Implements the Multiclass Tsetlin Machine, Convolutional Tsetlin Machine, Regression Tsetlin Machine, and Weighted Tsetlin Machine, with support for continuous features and multiple layers.',
   keywords ='pattern-recognition cuda machine-learning interpretable-machine-learning rule-based-machine-learning propositional-logic tsetlin-machine regression convolution classification multi-layer',
   packages=['tsetlinmachinecuda']
)



import numpy as np

from sklearn.preprocessing import FunctionTransformer
from sklearn.preprocessing import KBinsDiscretizer

import warnings
warnings.filterwarnings("ignore", category=UserWarning)

# From https://github.com/WojciechMigda/Tsetlini/blob/main/lib/examples/california-housing/src/produce_dataset_alt.py

def _as_bits(x, nbits):
	s = '1' * x + '0' * (nbits - x)
	return np.array([int(c) for c in s])

def _unpack_bits(a, nbits):
	if len(a.shape) > 2:
		raise ValueError("_unpack_bits: input array cannot have more than 2 dimensions, got {}".format(len(a.shape)))

	a = np.clip(a, 0, nbits)
	a_ = np.empty_like(a, dtype=np.uint64)
	np.rint(a, out=a_, casting='unsafe')
	F = np.frompyfunc(_as_bits, 2, 1)
	rv = np.stack(F(a_.ravel(), nbits)).reshape(a.shape[0], -1)
	return rv

class Booleanizer:
	def __init__(self,  max_bits_per_feature = 25):
		self.max_bits_per_feature = max_bits_per_feature

		self.kbd = KBinsDiscretizer(n_bins=max_bits_per_feature+1, encode='ordinal', strategy='quantile')

		return

	def fit(self, X):
		self.kbd_fitted = self.kbd.fit(X)
		
		return

	def transform(self, X):
		X_transformed = self.kbd_fitted.transform(X).astype(int)

		pre = FunctionTransformer(_unpack_bits, validate=False, kw_args={'nbits': self.max_bits_per_feature})
		return pre.fit_transform(X_transformed)


# Copyright (c) 2021 Ole-Christoffer Granmo

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This code implements the Convolutional Tsetlin Machine from paper arXiv:1905.09688
# https://arxiv.org/abs/1905.09688

code_header = """
	#include <curand_kernel.h>
	
	#define INT_SIZE 32

	#define LA_CHUNKS (((FEATURES-1)/INT_SIZE + 1))
	#define CLAUSE_CHUNKS ((CLAUSES-1)/INT_SIZE + 1)

	#if (FEATURES % 32 != 0)
	#define FILTER (~(0xffffffff << (FEATURES % INT_SIZE)))
	#else
	#define FILTER 0xffffffff
	#endif
"""

code_update = """
	extern "C"
    {
    	// Increment the states of each of those 32 Tsetlin Automata flagged in the active bit vector.
		__device__ inline void inc(unsigned int *ta_state, int clause, int chunk, unsigned int active)
		{
			unsigned int carry, carry_next;
			int id = clause*LA_CHUNKS*STATE_BITS + chunk*STATE_BITS;
			carry = active;
			for (int b = 0; b < STATE_BITS; ++b) {
				if (carry == 0)
					break;

				carry_next = ta_state[id + b] & carry; // Sets carry bits (overflow) passing on to next bit
				ta_state[id + b] = ta_state[id + b] ^ carry; // Performs increments with XOR
				carry = carry_next;
			}

			if (carry > 0) {
				for (int b = 0; b < STATE_BITS; ++b) {
					ta_state[id + b] |= carry;
				}
			}   
		}

		// Decrement the states of each of those 32 Tsetlin Automata flagged in the active bit vector.
		__device__ inline void dec(unsigned int *ta_state, int clause, int chunk, unsigned int active)
		{
			unsigned int carry, carry_next;
			int id = clause*LA_CHUNKS*STATE_BITS + chunk*STATE_BITS;
			carry = active;
			for (int b = 0; b < STATE_BITS; ++b) {
				if (carry == 0)
					break;
				carry_next = (~ta_state[id + b]) & carry; // Sets carry bits (overflow) passing on to next bit
				ta_state[id + b] = ta_state[id + b] ^ carry; // Performs increments with XOR
				carry = carry_next;
			}

			if (carry > 0) {
				for (int b = 0; b < STATE_BITS; ++b) {
					ta_state[id + b] &= ~carry;
				}
			} 
		}

		__device__ inline void calculate_clause_output(curandState *localState, unsigned int *ta_state, unsigned int *clause_output, int *clause_patch, unsigned int *drop_feature, int *X)
		{
			int output_one_patches[PATCHES];
			int output_one_patches_count;

			// Evaluate each patch (convolution)
			output_one_patches_count = 0;
			for (int patch = 0; patch < PATCHES; ++patch) {
				int patch_clause_output = 1;
				for (int la_chunk = 0; la_chunk < LA_CHUNKS-1; ++la_chunk) {
					if ((ta_state[la_chunk*STATE_BITS + STATE_BITS - 1] & (drop_feature[la_chunk] | X[patch*LA_CHUNKS + la_chunk])) != ta_state[la_chunk*STATE_BITS + STATE_BITS - 1]) {
						patch_clause_output = 0;
						break;
					}
				}

				if (((ta_state[(LA_CHUNKS-1)*STATE_BITS + STATE_BITS - 1] & (drop_feature[LA_CHUNKS-1] | X[patch*LA_CHUNKS + LA_CHUNKS - 1]) & FILTER) != (ta_state[(LA_CHUNKS-1)*STATE_BITS + STATE_BITS - 1] & FILTER))) {
					patch_clause_output = 0;
				}

				if (patch_clause_output) {
					output_one_patches[output_one_patches_count] = patch;
					output_one_patches_count++;
				}
			}
		
			if (output_one_patches_count > 0) {
				*clause_output = 1;
				int patch_id = curand(localState) % output_one_patches_count;
				*clause_patch = output_one_patches[patch_id];
			} else {
				*clause_output = 0;
				*clause_patch = -1;
			}
		}

		__device__ inline void update_clause(curandState *localState, int *clause_weight, unsigned int *ta_state, int clause_output, int clause_patch, unsigned int *drop_feature, int *X, int y, int class_sum)
		{
			int target = 1 - 2*(class_sum > y);
			
			if (target == -1 && curand_uniform(localState) > 1.0*Q/max(1, CLASSES-1)) {
				return;
			}

			int sign = (*clause_weight >= 0) - (*clause_weight < 0);
		
			int absolute_prediction_error = abs(y - class_sum);
			if (curand_uniform(localState) <= 1.0*absolute_prediction_error/(2*THRESHOLD)) {
				if (target*sign > 0) {
					#if WEIGHTED_CLAUSES == 1
						if (clause_output && abs(*clause_weight) < INT_MAX) {
							(*clause_weight) += sign;
						}
					#endif

					// Type I Feedback
					for (int la_chunk = 0; la_chunk < LA_CHUNKS; ++la_chunk) {
						// Generate random bit values
						unsigned int la_feedback = 0;
						for (int b = 0; b < INT_SIZE; ++b) {
							if (curand_uniform(localState) <= 1.0/S) {
								la_feedback |= (1 << b);
							}
						}

						if (clause_output) {
							#if BOOST_TRUE_POSITIVE_FEEDBACK == 1
								inc(ta_state, 0, la_chunk, (~drop_feature[la_chunk]) & X[clause_patch*LA_CHUNKS + la_chunk]);
							#else
								inc(ta_state, 0, la_chunk, (~drop_feature[la_chunk]) & X[clause_patch*LA_CHUNKS + la_chunk] & (~la_feedback));
							#endif

							dec(ta_state, 0, la_chunk, (~drop_feature[la_chunk]) & (~X[clause_patch*LA_CHUNKS + la_chunk]) & la_feedback);
						} else {
							dec(ta_state, 0, la_chunk, (~drop_feature[la_chunk]) & la_feedback);
						}
					}
				} else if (target*sign < 0 && clause_output) {
					// Type II Feedback

					#if WEIGHTED_CLAUSES == 1
						if (abs(*clause_weight) > 1) {
							(*clause_weight) -= sign;
						}
					#endif
					
					for (int la_chunk = 0; la_chunk < LA_CHUNKS; ++la_chunk) {
						inc(ta_state, 0, la_chunk, (~drop_feature[la_chunk]) & (~X[clause_patch*LA_CHUNKS + la_chunk]) & (~ta_state[la_chunk*STATE_BITS + STATE_BITS - 1]));
					}
				}
			}
		}

		// Evaluate example
		__global__ void evaluate(unsigned int *global_ta_state, int *clause_weights, int *class_sum, unsigned int *drop_clause, unsigned int *drop_feature, int *X, int example)
		{
			int index = blockIdx.x * blockDim.x + threadIdx.x;
			int stride = blockDim.x * gridDim.x;

			for (int clause = index; clause < CLAUSES; clause += stride) {
				if (drop_clause[clause] == 1) {
					continue;
				}

				unsigned int *ta_state = &global_ta_state[clause*LA_CHUNKS*STATE_BITS];

				int clause_output;
				for (int patch = 0; patch < PATCHES; ++patch) {
					clause_output = 1;
					for (int la_chunk = 0; la_chunk < LA_CHUNKS-1; ++la_chunk) {
						if ((ta_state[la_chunk*STATE_BITS + STATE_BITS - 1] & (drop_feature[la_chunk] | X[example*(LA_CHUNKS*PATCHES) + patch*LA_CHUNKS + la_chunk])) != ta_state[la_chunk*STATE_BITS + STATE_BITS - 1]) {
							clause_output = 0;
							break;
						}
					}

					if ((ta_state[(LA_CHUNKS-1)*STATE_BITS + STATE_BITS - 1] & (drop_feature[LA_CHUNKS-1] | X[example*(LA_CHUNKS*PATCHES) + patch*LA_CHUNKS + LA_CHUNKS-1]) & FILTER) != (ta_state[(LA_CHUNKS-1)*STATE_BITS + STATE_BITS - 1] & FILTER)) {
						clause_output = 0;
					}

					if (clause_output) {
						break;
					}
				}

				if (clause_output) {
					int class_id = clause / (CLAUSES / CLASSES);
					int clause_weight = clause_weights[clause];
					atomicAdd(&class_sum[class_id], clause_weight);					
				}
			}
		}

		// Update state of Tsetlin Automata team
		__global__ void update(curandState *state, unsigned int *global_ta_state, int *clause_weights, int *class_sum, unsigned int *drop_clause, unsigned int *drop_feature, int *X, int *y, int example)
		{
			int index = blockIdx.x * blockDim.x + threadIdx.x;
			int stride = blockDim.x * gridDim.x;

			/* Copy state to local memory for efficiency */  
			curandState localState = state[index];

			// Calculate clause output first
			for (unsigned long long clause = index; clause < CLAUSES; clause += stride) {
				if (drop_clause[clause] == 1) {
					continue;
				}
				
				int class_id = clause / (CLAUSES / CLASSES);

				unsigned int *ta_state = &global_ta_state[clause*LA_CHUNKS*STATE_BITS];

				unsigned int clause_output;
				int clause_patch;
				calculate_clause_output(&localState, ta_state, &clause_output, &clause_patch, drop_feature, &X[example*(LA_CHUNKS*PATCHES)]);

				int local_class_sum = class_sum[class_id];
				if (local_class_sum > THRESHOLD) {
					local_class_sum = THRESHOLD;
				} else if (local_class_sum < -THRESHOLD) {
					local_class_sum = -THRESHOLD;
				}
				update_clause(&localState, &clause_weights[clause], ta_state, clause_output, clause_patch, drop_feature, &X[example*(LA_CHUNKS*PATCHES)], y[example*CLASSES + class_id], local_class_sum);
			}
		
			state[index] = localState;
		}
    }
"""

code_evaluate = """
	extern "C"
    {
		// Evaluate examples
		__global__ void evaluate(unsigned int *global_ta_state, int *clause_weights, int *class_sum, int *X)
		{
			int index = blockIdx.x * blockDim.x + threadIdx.x;
			int stride = blockDim.x * gridDim.x;

			for (int clause = index; clause < CLAUSES; clause += stride) {
				unsigned int *ta_state = &global_ta_state[clause*LA_CHUNKS*STATE_BITS];

				int all_exclude = 1;
				for (int la_chunk = 0; la_chunk < LA_CHUNKS-1; ++la_chunk) {
					if (ta_state[la_chunk*STATE_BITS + STATE_BITS - 1] > 0) {
						all_exclude = 0;
						break;
					}
				}

				if ((ta_state[(LA_CHUNKS-1)*STATE_BITS + STATE_BITS - 1] & FILTER) > 0) {
					all_exclude = 0;
				}

				if (all_exclude) {
					continue;
				}

				for (int e = 0; e < NUMBER_OF_EXAMPLES; ++e) {
					int clause_output;
					for (int patch = 0; patch < PATCHES; ++patch) {
						clause_output = 1;
						for (int la_chunk = 0; la_chunk < LA_CHUNKS-1; ++la_chunk) {
							if ((ta_state[la_chunk*STATE_BITS + STATE_BITS - 1] & X[e*(LA_CHUNKS*PATCHES) + patch*LA_CHUNKS + la_chunk]) != ta_state[la_chunk*STATE_BITS + STATE_BITS - 1]) {
								clause_output = 0;
								break;
							}
						}

						if ((ta_state[(LA_CHUNKS-1)*STATE_BITS + STATE_BITS - 1] & X[e*(LA_CHUNKS*PATCHES) + patch*LA_CHUNKS + LA_CHUNKS-1] & FILTER) != (ta_state[(LA_CHUNKS-1)*STATE_BITS + STATE_BITS - 1] & FILTER)) {
							clause_output = 0;
						}

						if (clause_output) {
							break;
						}
					}

					if (clause_output) {
						int class_id = clause / (CLAUSES / CLASSES);

						int clause_weight = clause_weights[clause];
						atomicAdd(&class_sum[class_id*NUMBER_OF_EXAMPLES + e], clause_weight);
					}
				}
			}
		}
	}
"""

code_prepare = """
	extern "C"
    {
		__global__ void prepare(curandState *state, unsigned int *global_ta_state, int *clause_weights, int *class_sum)
		{
			int index = blockIdx.x * blockDim.x + threadIdx.x;
			int stride = blockDim.x * gridDim.x;

			curandState localState = state[index];

			for (unsigned long long clause = index; clause < CLAUSES; clause += stride) {
				#if NEGATIVE_CLAUSES == 1
					clause_weights[clause] = 1 - 2 * (clause % 2);
				#else
					clause_weights[clause] = 1;
				#endif
				
				unsigned int *ta_state = &global_ta_state[clause*LA_CHUNKS*STATE_BITS];
				for (int la_chunk = 0; la_chunk < LA_CHUNKS-1; ++la_chunk) {
					for (int b = 0; b < STATE_BITS-1; ++b) {
						ta_state[la_chunk*STATE_BITS + b] = ~0;
					}
					ta_state[la_chunk*STATE_BITS + STATE_BITS - 1] = 0;
				}
			}

			state[index] = localState;
		}
	}
"""

code_encode = """
	#include <curand_kernel.h>

	extern "C"
    {
		__global__ void prepare_encode(unsigned int *X, unsigned int *encoded_X, int number_of_examples, int dim_x, int dim_y, int dim_z, int patch_dim_x, int patch_dim_y, int append_negated, int class_features)
		{
			int index = blockIdx.x * blockDim.x + threadIdx.x;
			int stride = blockDim.x * gridDim.x;

			int number_of_features = class_features + patch_dim_x * patch_dim_y * dim_z + (dim_x - patch_dim_x) + (dim_y - patch_dim_y);
			int number_of_patches = (dim_x - patch_dim_x + 1) * (dim_y - patch_dim_y + 1);

			int number_of_ta_chunks;
			if (append_negated) {
				number_of_ta_chunks= (((2*number_of_features-1)/32 + 1));
			} else {
				number_of_ta_chunks= (((number_of_features-1)/32 + 1));
			}

			for (int i = index; i < number_of_examples * number_of_patches * number_of_ta_chunks; i += stride) {
				encoded_X[i] = 0;
			}
		}
	
		__global__ void encode(unsigned int *X, unsigned int *encoded_X, int number_of_examples, int dim_x, int dim_y, int dim_z, int patch_dim_x, int patch_dim_y, int append_negated, int class_features)
		{
			int index = blockIdx.x * blockDim.x + threadIdx.x;
			int stride = blockDim.x * gridDim.x;

			int global_number_of_features = dim_x * dim_y * dim_z;
			int number_of_features = class_features + patch_dim_x * patch_dim_y * dim_z + (dim_x - patch_dim_x) + (dim_y - patch_dim_y);
			int number_of_patches = (dim_x - patch_dim_x + 1) * (dim_y - patch_dim_y + 1);

			int number_of_ta_chunks;
			if (append_negated) {
				number_of_ta_chunks= (((2*number_of_features-1)/32 + 1));
			} else {
				number_of_ta_chunks= (((number_of_features-1)/32 + 1));
			}

			unsigned int *Xi;
			unsigned int *encoded_Xi;

			unsigned int input_step_size = global_number_of_features;

			for (int i = index; i < number_of_examples; i += stride) {
				unsigned int encoded_pos = i * number_of_patches * number_of_ta_chunks;
				unsigned int input_pos = i * input_step_size;

				int patch_nr = 0;
				// Produce the patches of the current image
				for (int y = 0; y < dim_y - patch_dim_y + 1; ++y) {
					for (int x = 0; x < dim_x - patch_dim_x + 1; ++x) {
						Xi = &X[input_pos];
						encoded_Xi = &encoded_X[encoded_pos];

						// Encode class into feature vector 
						for (int class_feature = 0; class_feature < class_features; ++class_feature) {

							int chunk_nr = (class_feature + number_of_features) / 32;
							int chunk_pos = (class_feature + number_of_features) % 32;
							encoded_Xi[chunk_nr] |= (1 << chunk_pos);
						}

						// Encode y coordinate of patch into feature vector 
						for (int y_threshold = 0; y_threshold < dim_y - patch_dim_y; ++y_threshold) {
							int patch_pos = class_features + y_threshold;

							if (y > y_threshold) {
								int chunk_nr = patch_pos / 32;
								int chunk_pos = patch_pos % 32;
								encoded_Xi[chunk_nr] |= (1 << chunk_pos);
							} else if (append_negated) {
								int chunk_nr = (patch_pos + number_of_features) / 32;
								int chunk_pos = (patch_pos + number_of_features) % 32;
								encoded_Xi[chunk_nr] |= (1 << chunk_pos);
							}
						}

						// Encode x coordinate of patch into feature vector
						for (int x_threshold = 0; x_threshold < dim_x - patch_dim_x; ++x_threshold) {
							int patch_pos = class_features + (dim_y - patch_dim_y) + x_threshold;

							if (x > x_threshold) {
								int chunk_nr = patch_pos / 32;
								int chunk_pos = patch_pos % 32;

								encoded_Xi[chunk_nr] |= (1 << chunk_pos);
							} else if (append_negated) {
								int chunk_nr = (patch_pos + number_of_features) / 32;
								int chunk_pos = (patch_pos + number_of_features) % 32;
								encoded_Xi[chunk_nr] |= (1 << chunk_pos);
							}
						} 

						// Encode patch content into feature vector
						for (int p_y = 0; p_y < patch_dim_y; ++p_y) {
							for (int p_x = 0; p_x < patch_dim_x; ++p_x) {
								for (int z = 0; z < dim_z; ++z) {
									int image_pos = (y + p_y)*dim_x*dim_z + (x + p_x)*dim_z + z;
									int patch_pos = class_features + (dim_y - patch_dim_y) + (dim_x - patch_dim_x) + p_y * patch_dim_x * dim_z + p_x * dim_z + z;

									if (Xi[image_pos] == 1) {
										int chunk_nr = patch_pos / 32;
										int chunk_pos = patch_pos % 32;
										encoded_Xi[chunk_nr] |= (1 << chunk_pos);
									} else if (append_negated) {
										int chunk_nr = (patch_pos + number_of_features) / 32;
										int chunk_pos = (patch_pos + number_of_features) % 32;
										encoded_Xi[chunk_nr] |= (1 << chunk_pos);
									}
								}
							}
						}
						encoded_pos += number_of_ta_chunks;
						patch_nr++;
					}
				}
			}
		}
	}
"""

code_transform = """
	extern "C"
    {
		// Transform examples
		__global__ void transform(unsigned int *global_ta_state, int *X, int *transformed_X)
		{
			int index = blockIdx.x * blockDim.x + threadIdx.x;
			int stride = blockDim.x * gridDim.x;

			for (int clause = index; clause < CLAUSES; clause += stride) {
				unsigned int *ta_state = &global_ta_state[clause*LA_CHUNKS*STATE_BITS];

				int all_exclude = 1;
				for (int la_chunk = 0; la_chunk < LA_CHUNKS-1; ++la_chunk) {
					if (ta_state[la_chunk*STATE_BITS + STATE_BITS - 1] > 0) {
						all_exclude = 0;
						break;
					}
				}

				if ((ta_state[(LA_CHUNKS-1)*STATE_BITS + STATE_BITS - 1] & FILTER) > 0) {
					all_exclude = 0;
				}

				if (all_exclude) {
					for (unsigned long long e = 0; e < NUMBER_OF_EXAMPLES; ++e) {
						transformed_X[e*CLAUSES + clause] = 0;
					}
					
					continue;
				}

				for (unsigned long long e = 0; e < NUMBER_OF_EXAMPLES; ++e) {
					int clause_output;
					for (int patch = 0; patch < PATCHES; ++patch) {
						clause_output = 1;
						for (int la_chunk = 0; la_chunk < LA_CHUNKS-1; ++la_chunk) {
							if ((ta_state[la_chunk*STATE_BITS + STATE_BITS - 1] & X[e*(LA_CHUNKS*PATCHES) + patch*LA_CHUNKS + la_chunk]) != ta_state[la_chunk*STATE_BITS + STATE_BITS - 1]) {
								clause_output = 0;
								break;
							}
						}

						if ((ta_state[(LA_CHUNKS-1)*STATE_BITS + STATE_BITS - 1] & X[e*(LA_CHUNKS*PATCHES) + patch*LA_CHUNKS + LA_CHUNKS-1] & FILTER) != (ta_state[(LA_CHUNKS-1)*STATE_BITS + STATE_BITS - 1] & FILTER)) {
							clause_output = 0;
						}

						if (clause_output) {
							break;
						}
					}

					transformed_X[e*CLAUSES + clause] = clause_output;
				}
			}
		}
	}
"""
